from typing import List

from qulacs import DensityMatrix, QuantumState, observable

from ouqu_tp.internal.auto_noise import auto_evo_noise, auto_noise
from ouqu_tp.internal.QASMtoqulacs import QASM_to_qulacs


def getval_do(input_strs: List[str], ferfile: str) -> float:
    testcircuit = QASM_to_qulacs(input_strs, remap_remove=True)

    out_state = QuantumState(testcircuit.get_qubit_count())
    testcircuit.update_quantum_state(out_state)

    obs = observable.create_observable_from_openfermion_file(ferfile)
    return float(obs.get_expectation_value(out_state))
    # qulacsの型アノテーションないので、怒りのキャスト


def getval_noise_do(
    input_strs: List[str], ferfile: str, p1: float, p2: float, pm: float, pp: float
) -> float:

    precircuit = QASM_to_qulacs(input_strs, remap_remove=True)
    testcircuit = auto_noise(precircuit, p1, p2, pm, pp)

    out_state = DensityMatrix(testcircuit.get_qubit_count())
    testcircuit.update_quantum_state(out_state)

    obs = observable.create_observable_from_openfermion_file(ferfile)
    return float(obs.get_expectation_value(out_state))
    # qulacsの型アノテーションないので、怒りのキャスト


def getval_noiseevo_do(
    input_strs: List[str],
    ferfile: str,
    shots: int,
    dt: float,
    OZ: float,
    OX: float,
    ORes: float,
    decay_rate_ph: float,
    decay_rate_amp: float,
    evodt: float,
) -> float:
    precircuit = QASM_to_qulacs(input_strs, remap_remove=True)
    testcircuit = auto_evo_noise(
        precircuit, dt, OZ, OX, ORes, decay_rate_ph, decay_rate_amp, evodt
    )
    # print(testcircuit)
    ans = 0.0
    obs = observable.create_observable_from_openfermion_file(ferfile)
    for _ in range(shots):
        out_state = QuantumState(testcircuit.get_qubit_count())
        testcircuit.update_quantum_state(out_state)
        ans += float(obs.get_expectation_value(out_state)) / shots
    return ans



import subprocess
from typing import List

import typer

from ouqu_tp.internal.getval import getval_noise_do
from ouqu_tp.internal.sampleval import sampleval_noise_do
from ouqu_tp.internal.simulate import simulate_noise_do

app = typer.Typer()


@app.command("getval")
def getval_noisy_call(
    input_qasm_file: str = typer.Option(...),
    input_openfermion_file: str = typer.Option(...),
    p1: float = 0,
    p2: float = 0,
    pm: float = 0,
    pp: float = 0,
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(["staq", "-m", "--evaluate-all", input_qasm_file])
            .decode()
            .splitlines()
        )

    print(getval_noise_do(cpl_qasm, input_openfermion_file, p1, p2, pm, pp))


@app.command("sampleval")
def sampleval_noisy_call(
    input_qasm_file: str = typer.Option(...),
    input_openfermion_file: str = typer.Option(...),
    shots: int = typer.Option(...),
    p1: float = 0,
    p2: float = 0,
    pm: float = 0,
    pp: float = 0,
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(["staq", "-m", "--evaluate-all", input_qasm_file])
            .decode()
            .splitlines()
        )
    print(sampleval_noise_do(cpl_qasm, input_openfermion_file, shots, p1, p2, pm, pp))


@app.command("simulate")
def simulate_noisy_call(
    input_qasm_file: str = typer.Option(...),
    shots: int = typer.Option(...),
    p1: float = 0,
    p2: float = 0,
    pm: float = 0,
    pp: float = 0,
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(["staq", "-m", "--evaluate-all", input_qasm_file])
            .decode()
            .splitlines()
        )
    (kekka, n_qubit) = simulate_noise_do(cpl_qasm, shots, p1, p2, pm, pp)
    # input_listを直接ぶち込む
    for aaaa in kekka:
        moziretu = "{:b}".format(aaaa)
        while len(moziretu) < n_qubit:
            moziretu = "0" + moziretu
        print(moziretu)


if __name__ == "__main__":
    app()



from dataclasses import dataclass
from logging import getLogger
from typing import Any, Dict, Optional

from ouqu_tp.servicers.default_device_topology_json import (
    DEFAULT_DEVICE_TOPOLOGY_JSON,
)
from ouqu_tp.servicers.exec_ouqu import ouqu


@dataclass
class TranspileResponse:
    status: str
    message: str
    qasm: str
    qubit_mapping: Dict[int, int]


class TranspilerService(object):
    def transpile(self, qasm: str, device_topology_json: Optional[str] = None) -> Any:
        logger = getLogger(__name__)
        try:
            logger.debug(f"qasm: {qasm}")
            logger.debug(f"device_topology_json: {device_topology_json}")

            cn_json = ""
            # リクエストで渡されたJSONが優先
            if device_topology_json is None:
                cn_json = DEFAULT_DEVICE_TOPOLOGY_JSON
            else:
                cn_json = device_topology_json

            logger.debug(f"ouqu cn_json: {cn_json}")

            result, mapping = ouqu(qasm, cn_json)
            logger.debug(f"ouqu result: {result}")
            logger.debug(f"ouqu mapping: {mapping}")
            return TranspileResponse(
                status="SUCCESS", message="", qasm=result, qubit_mapping=mapping
            )
        except Exception as e:
            logger.exception(e)
            return TranspileResponse(
                status="FAILURE", message=str(e), qasm="", qubit_mapping={}
            )



import subprocess
import sys
from typing import List

import numpy as np
import typer

from ouqu_tp.internal.make_Cnet import make_Cnet_put
from ouqu_tp.internal.trancepile import (
    trance_do,
    trance_pulse_do,
    trance_res_do,
)

app = typer.Typer()


@app.command("trance")
def trance_call(
    input_qasm_file: str = typer.Option(...),
    input_cnot_json_file: str = typer.Option(...),
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(
                [
                    "staq",
                    "-S",
                    "-O2",
                    "-m",
                    "-d",
                    input_cnot_json_file,
                    "--evaluate-all",
                    input_qasm_file,
                ]
            )
            .decode()
            .splitlines()
        )
    out_QASM = trance_do(cpl_qasm)
    for cpl_qasm_comment_line in cpl_qasm:
        if cpl_qasm_comment_line[0:2] == "//":
            print(cpl_qasm_comment_line)
            # コメントの垂れ流しを行います
    for QASM_line in out_QASM:
        print(QASM_line)


@app.command("trance_res")
def trance_res_call(
    input_qasm_file: str = typer.Option(...),
    input_cnot_json_file: str = typer.Option(...),
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(
                [
                    "staq",
                    "-S",
                    "-O2",
                    "-m",
                    "-d",
                    input_cnot_json_file,
                    "--evaluate-all",
                    input_qasm_file,
                ]
            )
            .decode()
            .splitlines()
        )
    out_QASM = trance_res_do(cpl_qasm)
    for cpl_qasm_comment_line in cpl_qasm:
        if cpl_qasm_comment_line[0:2] == "//":
            print(cpl_qasm_comment_line)
            # コメントの垂れ流しを行います
    for QASM_line in out_QASM:
        print(QASM_line)


@app.command("trance_pulse")
def trance_pulse_call(
    input_qasm_file: str = typer.Option(...),
    input_cnot_json_file: str = typer.Option(...),
    cnot_net_file: str = typer.Option(...),
    dt: float = 0.01,
    OZ: float = 10,
    OX: float = 10,
    ORes: float = 1,
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(
                [
                    "staq",
                    "-S",
                    "-O2",
                    "-m",
                    "-d",
                    input_cnot_json_file,
                    "--evaluate-all",
                    input_qasm_file,
                ]
            )
            .decode()
            .splitlines()
        )
    ff = open(cnot_net_file, "r")
    Cnet_list = ff.readlines()
    result_array = trance_pulse_do(cpl_qasm, Cnet_list, dt, OZ, OX, ORes, 0)
    np.set_printoptions(threshold=99999999)
    """
    for cpl_qasm_comment_line in cpl_qasm:
        if cpl_qasm_comment_line[0:2] == "//":
            print(cpl_qasm_comment_line)
            # コメントの垂れ流しを行います
    """
    np.savetxt(sys.stdout.buffer, result_array)
    # savetxt で出力したかったからこうなった


@app.command("makeCnet")
def makeCnet_call(
    cnot_net_file: str = typer.Option(...),
) -> None:

    ff = open(cnot_net_file, "r")
    Cnet_list = ff.readlines()
    make_Cnet_put(Cnet_list)


if __name__ == "__main__":
    app()



import logging
import subprocess
from xmlrpc.client import Boolean

import typer

import ouqu_tp.cli.ideal as ideal
import ouqu_tp.cli.noisy as noisy
import ouqu_tp.cli.trance as trance

app = typer.Typer()
app.add_typer(ideal.app, name="ideal")
app.add_typer(noisy.app, name="noisy")
app.add_typer(trance.app, name="trance")


def is_staq_installed() -> Boolean:
    try:
        subprocess.run(["staq", "--help"], stdout=subprocess.DEVNULL, check=True)
        return True
    except Exception:
        return False


@app.callback()
def debugflag(debug: bool = False) -> None:
    """
    Debug mode for output log files.
    """
    if debug:
        logger = logging.getLogger()
        formatter = logging.Formatter(
            "%(asctime)s %(name)s %(funcName)s [%(levelname)s]: %(message)s"
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)


def main() -> None:
    if is_staq_installed() is False:
        raise Exception(
            "[ERROR] staq seems to be not installed.\n"
            "please install staq from https://github.com/softwareQinc/staq"
        )

    app()


if __name__ == "__main__":
    main()



from qiskit import QuantumCircuit

input_circuit = QuantumCircuit.from_qasm_file("sample/grover_moto.qasm")
input_circuit.draw(output="mpl").savefig("sample/grover_moto._graph.png")
output_circuit = QuantumCircuit.from_qasm_file("sample/grover_cpl.qasm")
output_circuit.draw(output="mpl").savefig("sample/grover_cpl_graph.png")



import os
import re
import subprocess
import tempfile
from typing import Dict, Tuple

from parse import parse

from ouqu_tp.servicers import config

DELETE_FLG = config.get_config_DELETE_TMP_FLG()
ADD_INC_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "add.inc")


def add_gate_def(qasm_str: str) -> str:
    result = ""

    def read_add() -> str:
        with open(ADD_INC_PATH) as f:
            s = f.read()
            return s

    for line in qasm_str.split("\n"):
        result = result + line + "\n"
        if "qelib1.inc" in line.lower():
            result = result + read_add() + "\n"

    return result


def convert_qubit_to_qreg(qasm: str) -> str:
    result = ""
    q_str = "q"
    c_str = "c"
    n_size = ""
    is_exist_bit = False
    for line in qasm.split("\n"):
        # 前スペースを取る
        line = line.strip()
        # qubit -> qreg
        ary = parse("qubit[{n}] {v};", line)
        if ary is not None:
            n = ary["n"]
            n_size = n
            v = ary["v"].strip()
            q_str = v
            q = f"qreg {v}[{n}];"
            result = result + q + "\n"
        else:
            # bit -> creg
            ary = parse("bit[{n}] {v};", line)
            if ary is not None:
                is_exist_bit = True
                n = ary["n"]
                n_size = n
                v = ary["v"].strip()
                c_str = v
                q = f"creg {v}[{n}];"
                result = result + q + "\n"
            else:
                result = result + line + "\n"

        # cregが存在する場合
        ary = parse("creg {v}[{n}];", line)
        if ary is not None:
            is_exist_bit = True

    if not is_exist_bit:
        result = result + f"creg c[{n_size}];"

    if "measure" in qasm.lower():
        return result
    else:
        str = result + f"measure {q_str} -> {c_str};"
        return str


def convert_v3_to_v2_measure(qasm: str) -> str:
    result = ""
    for line in qasm.split("\n"):
        # `c = measure q;` -> `measure q -> c;`
        ary = parse("{vc}= measure {vq};", line)
        if ary is not None and ary["vc"][0:2] != "//":
            vc = ary["vc"].strip()
            vq = ary["vq"].strip()
            q = f"measure {vq}->{vc};"
            result = result + q + "\n"
        else:
            result = result + line + "\n"

    return result


def convert_v3_to_v2(qasm: str) -> str:
    str = re.sub(r"OPENQASM\s+3.0;", "OPENQASM 2.0;", qasm, flags=re.IGNORECASE)
    str = re.sub(r"OPENQASM\s+3;", "OPENQASM 2.0;", str, flags=re.IGNORECASE)
    if "stdgates.inc" not in qasm:
        str = re.sub(
            "OPENQASM 2.0;",
            'OPENQASM 2.0;\ninclude "stdgates.inc";\n',
            str,
            flags=re.IGNORECASE,
        )
    str = re.sub("stdgates.inc", "qelib1.inc", str, flags=re.IGNORECASE)

    str = convert_qubit_to_qreg(str)
    str = convert_v3_to_v2_measure(str)
    return str


def convert_v2_to_v3_measure(qasm: str) -> str:
    result = ""
    for line in qasm.split("\n"):
        # `measure q -> c;` -> `c = measure q;`
        ary = parse("measure{vq}->{vc};", line)
        if ary is not None:
            vc = ary["vc"].strip()
            vq = ary["vq"].strip()
            q = f"{vc} = measure {vq};"
            result = result + q + "\n"
        else:
            result = result + line + "\n"

    return result


def convert_v2_to_v3(qasm: str) -> str:
    str = re.sub("OPENQASM 2.0;", "OPENQASM 3.0;", qasm, flags=re.IGNORECASE)
    str = re.sub("qelib1.inc", "stdgates.inc", str, flags=re.IGNORECASE)
    str = convert_v2_to_v3_measure(str)
    return str


def add_newline_to_semicolons(qasm_str: str) -> str:
    # 正規表現でセミコロンに続く改行がない場合に改行を追加する
    pattern = r";(?!\n)"
    repl = ";\n"
    result = re.sub(pattern, repl, qasm_str)
    return result


def exec_ouqu(qasm_str: str, device_topology_json: str) -> str:
    device_topology_json = device_topology_json.replace("device_id", "name")
    with tempfile.NamedTemporaryFile(mode="w+t", delete=DELETE_FLG) as fcnot:
        fcnot.write(device_topology_json)
        fcnot.seek(0)

        with tempfile.NamedTemporaryFile(mode="w+t", delete=DELETE_FLG) as fqasm:
            fqasm.write(qasm_str)
            fqasm.seek(0)

            cmd = f"ouqu-tp trance trance --input-qasm-file={fqasm.name} --input-cnot-json-file={fcnot.name}"
            subprocess_output = subprocess.run(
                cmd, capture_output=True, text=True, shell=True
            )
            # TODO:swapでなぜか、staqが"Error: linear operator is not invertible"を標準エラーに出力するので対応
            if len(subprocess_output.stderr) > 0:
                if (
                    "Error: linear operator is not invertible"
                    in subprocess_output.stderr
                ):
                    return subprocess_output.stdout
                else:
                    raise Exception(subprocess_output.stderr)
            return subprocess_output.stdout


def get_mapping(qasm: str) -> Dict[int, int]:
    mapping = {}

    for line in qasm.split("\n"):
        instr = line.lower().strip().replace(" ", "").replace("\t", "")
        if instr[0:4] == "//q[":
            ary = parse("//q[{:d}]-->q[{:d}]", instr)
            if ary is not None:
                mapping[ary[0]] = ary[1]

    return mapping


def ouqu(qasm_str: str, device_topology_json: str) -> Tuple[str, Dict[int, int]]:
    qasm_str = add_newline_to_semicolons(qasm_str)
    qasm_str = convert_v3_to_v2(qasm_str)
    qasm_str = add_gate_def(qasm_str)
    result = exec_ouqu(qasm_str, device_topology_json)
    result = convert_v2_to_v3(result)
    mapping = get_mapping(result)
    return result, mapping



from typing import List, Tuple

from qulacs import NoiseSimulator, QuantumState

from ouqu_tp.internal.auto_noise import auto_evo_noise, auto_noise
from ouqu_tp.internal.QASMtoqulacs import QASM_to_qulacs


def simulate_do(input_strs: List[str], shots: int) -> Tuple[List[int], int]:
    testcircuit = QASM_to_qulacs(input_strs, remap_remove=True)
    n_qubit = testcircuit.get_qubit_count()

    out_state = QuantumState(n_qubit)
    testcircuit.update_quantum_state(out_state)
    kekka = out_state.sampling(shots)
    return (kekka, n_qubit)


def simulate_noise_do(
    input_strs: List[str], shots: int, p1: float, p2: float, pm: float, pp: float
) -> Tuple[List[int], int]:
    precircuit = QASM_to_qulacs(input_strs, remap_remove=True)
    testcircuit = auto_noise(precircuit, p1, p2, pm, pp)
    n_qubit = testcircuit.get_qubit_count()

    out_state = QuantumState(n_qubit)
    nsim = NoiseSimulator(testcircuit, out_state)

    kekka = nsim.execute(shots)
    return (kekka, n_qubit)


def simulate_noiseevo_do(
    input_strs: List[str],
    shots: int,
    dt: float,
    OZ: float,
    OX: float,
    ORes: float,
    decay_rate_ph: float,
    decay_rate_amp: float,
    evodt: float,
) -> Tuple[List[int], int]:
    precircuit = QASM_to_qulacs(input_strs, remap_remove=True)
    testcircuit = auto_evo_noise(
        precircuit, dt, OZ, OX, ORes, decay_rate_ph, decay_rate_amp, evodt
    )
    n_qubit = testcircuit.get_qubit_count()

    out_state = QuantumState(n_qubit)
    kekka = []
    for _ in range(shots):
        out_state.set_zero_state()
        testcircuit.update_quantum_state(out_state)
        kekka.append(out_state.sampling(1)[0])
    return (kekka, n_qubit)



from typing import List

from qulacs import observable

from ouqu_tp.internal.auto_noise import auto_evo_noise, auto_noise
from ouqu_tp.internal.QASMtoqulacs import QASM_to_qulacs
from ouqu_tp.internal.shot_obs import (
    get_measurement,
    get_noise_meseurment,
    get_noiseevo_meseurment,
)


def sampleval_do(input_strs: List[str], ferfile: str, shots: int) -> float:
    testcircuit = QASM_to_qulacs(input_strs, remap_remove=True)

    obs = observable.create_observable_from_openfermion_file(ferfile)

    return get_measurement(testcircuit, obs, shots)


def sampleval_noise_do(
    input_strs: List[str],
    ferfile: str,
    shots: int,
    p1: float,
    p2: float,
    pm: float,
    pp: float,
) -> float:
    precircuit = QASM_to_qulacs(input_strs, remap_remove=True)
    testcircuit = auto_noise(precircuit, p1, p2, pm, pp)

    obs = observable.create_observable_from_openfermion_file(ferfile)
    return get_noise_meseurment(testcircuit, obs, shots)


def sampleval_noiseevo_do(
    input_strs: List[str],
    ferfile: str,
    shots: int,
    dt: float,
    OZ: float,
    OX: float,
    ORes: float,
    decay_rate_ph: float,
    decay_rate_amp: float,
    evodt: float,
) -> float:
    precircuit = QASM_to_qulacs(input_strs, remap_remove=True)
    testcircuit = auto_evo_noise(
        precircuit, dt, OZ, OX, ORes, decay_rate_ph, decay_rate_amp, evodt
    )

    obs = observable.create_observable_from_openfermion_file(ferfile)
    return get_noiseevo_meseurment(testcircuit, obs, shots)



from os import getenv
from typing import Any


def get_config(key: str, default_value: Any) -> Any:
    return getenv(key, default_value)


# ログ


def get_config_LOG_FORMAT(default_value: str) -> str:
    return str(get_config("LOG_FORMAT", default_value))


def get_config_LOG_LEVEL(default_value: int) -> int:
    return int(get_config("LOG_LEVEL", default_value))


def get_config_LOG_PATH(default_value: str) -> str:
    return str(get_config("LOG_PATH", default_value))


# PORT


def get_config_PORT(default_value: int) -> int:
    return int(get_config("PORT", default_value))


# 一時ファイルの削除フラグ: Trueの場合、削除する


def get_config_DELETE_TMP_FLG() -> bool:
    return str(get_config("DELETE_TMP_FLG", "FALSE")).upper() == "TRUE"



from qiskit import Aer, QuantumCircuit, execute

input_circuit = QuantumCircuit.from_qasm_file("sample/input.qasm")
input_circuit.draw(output="mpl").savefig("sample/input_qasm_graph.png")
output_circuit = QuantumCircuit.from_qasm_file("sample/output.qasm")
output_circuit.draw(output="mpl").savefig("sample/output_qasm_graph.png")

vector_sim = Aer.get_backend("statevector_simulator")
job = execute(input_circuit, vector_sim)
ket = job.result().get_statevector()
for amplitude in ket:
    print(amplitude)

print()

job2 = execute(output_circuit, vector_sim)
ket2 = job2.result().get_statevector()

bit_narabe = [0, 2, 1, 5, 4]

naiseki = 0
for i in range(32):
    bit_ter = 0
    for j in range(5):
        if (i & (2**j)) > 0:
            bit_ter += 2 ** bit_narabe[j]

    print(ket2[bit_ter])

    naiseki += ket[i] * ket2[bit_ter].conj()

print(abs(naiseki))



import json
import re
import typing
from cmath import phase
from logging import NullHandler, getLogger
from typing import List

import numpy as np
from parse import parse, search
from qulacs import QuantumCircuit, QuantumGateBase, QuantumState
from qulacs.gate import DenseMatrix, Identity, Measurement

from ouqu_tp.internal.tran import (
    CRes,
    CResdag,
    check_is_CRes,
    check_is_CResdag,
)

logger = getLogger(__name__)
logger.addHandler(NullHandler())


def can_get_dence(gate: QuantumGateBase) -> bool:
    # dencematrixにできるかの関数
    return True


# qelib1.inc にあるゲートのほとんどにqulacsを対応させる
# sqrtYはない
# まずは、qulacs to QASM


def qulacs_to_QASM(cir: QuantumCircuit) -> typing.List[str]:
    # QASM風です
    #

    out_strs = [
        "OPENQASM 2.0;",
        'include "qelib1.inc";',
        f"qreg q[{cir.get_qubit_count()}];",
        "creg c[__len_creg_idx__];",
    ]

    creg_idx = list()

    for kai in range(cir.get_gate_count()):
        it = cir.get_gate(kai)
        clis = it.get_control_index_list()
        tlis = it.get_target_index_list()

        # print(f"it.get_name(): {it.get_name()}")

        if it.get_name() == "CNOT":
            out_strs.append(f"cx q[{clis[0]}],q[{tlis[0]}];")
        elif it.get_name() == "CZ":
            out_strs.append(f"cz q[{clis[0]}],q[{tlis[0]}];")
        elif it.get_name() == "SWAP":
            out_strs.append(f"swap q[{tlis[0]}],q[{tlis[1]}];")
        elif it.get_name() == "Identity":
            out_strs.append(f"id q[{tlis[0]}];")
        elif it.get_name() == "X":
            out_strs.append(f"x q[{tlis[0]}];")
        elif it.get_name() == "Y":
            out_strs.append(f"y q[{tlis[0]}];")
        elif it.get_name() == "Z":
            out_strs.append(f"z q[{tlis[0]}];")
        elif it.get_name() == "H":
            out_strs.append(f"h q[{tlis[0]}];")
        elif it.get_name() == "S":
            out_strs.append(f"s q[{tlis[0]}];")
        elif it.get_name() == "Sdag":
            out_strs.append(f"sdg q[{tlis[0]}];")
        elif it.get_name() == "T":
            out_strs.append(f"t q[{tlis[0]}];")
        elif it.get_name() == "Tdag":
            out_strs.append(f"tdg q[{tlis[0]}];")
        elif it.get_name() == "sqrtX":
            out_strs.append(f"sx q[{tlis[0]}];")
        elif it.get_name() == "sqrtXdag":
            out_strs.append(f"sxdg q[{tlis[0]}];")
        elif it.get_name() == "X-rotation":
            matrix = it.get_matrix()
            angle = phase(matrix[0][0] - matrix[1][0]) * 2
            if abs(angle) > 1e-5:
                out_strs.append(f"rx({angle}) q[{tlis[0]}];")
        elif it.get_name() == "Y-rotation":
            matrix = it.get_matrix()
            angle = phase(matrix[0][0] + matrix[1][0] * 1.0j) * 2
            if abs(angle) > 1e-5:
                out_strs.append(f"ry({angle}) q[{tlis[0]}];")
        elif it.get_name() == "Z-rotation":
            matrix = it.get_matrix()
            angle = phase(matrix[1][1] / matrix[0][0])
            if abs(angle) > 1e-5:
                out_strs.append(f"rz({angle}) q[{tlis[0]}];")
        elif it.get_name() == "CPTP":
            target_idx_list = it.get_target_index_list()
            if len(target_idx_list) > 0:
                target_idx = target_idx_list[0]
                measure_dict = json.loads(it.to_json())
                if "classical_register_address" in measure_dict:
                    classical_idx = measure_dict["classical_register_address"]
                    out_strs.append(f"measure q[{target_idx}]->c[{classical_idx}];")
                    creg_idx.append(int(classical_idx))
                else:
                    out_strs.append(f"measure q[{target_idx}]->c[{target_idx}];")
                    creg_idx.append(int(target_idx))
        elif check_is_CRes(it):
            out_strs.append(f"CRes q[{tlis[0]}],q[{tlis[1]}];")
        elif check_is_CResdag(it):
            out_strs.append(f"CResdag q[{tlis[0]}],q[{tlis[1]}];")
        elif can_get_dence(it):
            now_string = ""
            bit = len(it.get_target_index_list())
            matrix = it.get_matrix()
            now_string += "DenseMatrix("
            now_string += str(len(it.get_target_index_list()))
            now_string += ","
            now_string += str(len(it.get_control_index_list()))
            for i in range(2**bit):
                for j in range(2**bit):
                    now_string += f",{matrix[i][j].real:.6g},{matrix[i][j].imag:.6g}"
            for i in range(len(it.get_control_index_list())):
                if it.get_control_value_list()[i] == 1:
                    now_string += ",1"
                else:
                    now_string += ",0"

            now_string += ") "
            for aaa in it.get_target_index_list():
                if aaa == it.get_target_index_list()[0]:
                    now_string += f"q[{aaa}]"
                else:
                    now_string += f",q[{aaa}]"
            for aaa in it.get_control_index_list():
                now_string += f",q[{aaa}]"
            now_string += ";"
            out_strs.append(now_string)
        else:
            out_strs.append("unknown gate:" + it.get_name())
            logger.warning("unknown gate:" + it.get_name())
            print(f"unknown gate: {it.get_name()}")
        # get_matrix が効かないゲートの対応
        # 1qubitのDenseMatrixはu3ゲートに直すべきだが、やってない

    # rewite creg
    creg_idx.append(0)
    num_creg = max(creg_idx) + 1
    out_strs[3] = re.sub(r"__len_creg_idx__", str(num_creg), out_strs[3])
    return out_strs


"""
  (0.707107,0)          (0,0) (-0,-0.707107)          (0,0)
         (0,0)   (0.707107,0)          (0,0)   (0,0.707107)
(-0,-0.707107)          (0,0)   (0.707107,0)          (0,0)
         (0,0)   (0,0.707107)          (0,0)   (0.707107,0)

DenseMatrix(2,1,0.707107,0,0,0,-0,-0.707107,0,0,0,0,0.707107,0,0,0,0,0.707107,-0,-0.707107,0,0,0.707107,0,0,0,0,0,0,0.707107,0,0,0.707107,0,1) q[0],q[1],q[2];

"""


def QASM_to_qulacs(
    input_strs: typing.List[str], *, remap_remove: bool = False
) -> QuantumCircuit:
    # 仕様: キュービットレジスタはq[]のやつだけにしてください
    n = 0
    mapping: List[int] = []

    for instr_moto in input_strs:
        instr = instr_moto.lower().strip().replace(" ", "").replace("\t", "")
        # 全部小文字にして、前後の改行を削除、　すべての空白とタブを削除した状態でマッチングします
        if instr[0:4] == "qreg":
            ary = parse("qreg{v}[{n}];", instr)
            if ary is not None:
                n = int(ary["n"].strip())
                # v = ary["v"].strip()
                cir = QuantumCircuit(n)
                if len(mapping) == 0:
                    mapping = list(range(n))
        elif instr[0:4] == "creg":
            continue
        elif instr[0:4] == "gate":
            continue
        elif remap_remove and instr[0:4] == "//q[":
            ary = parse("//q[{:d}]-->q[{:d}]", instr)
            if ary is not None:
                mapping[ary[0]] = ary[1]
        elif remap_remove and instr[0:8] == "//qubits":
            ary = parse("//qubits:{:d}", instr)
            mapping = list(range(ary[0]))
        elif instr[0:2] == "//":
            continue
        elif instr.lower() == "openqasm2.0;":
            continue
        elif instr.lower() == 'include"qelib1.inc";':
            continue
        elif instr.isspace():
            continue
        elif instr == "}":
            continue
        elif instr == "":
            continue

        elif instr[0:2] == "cx":
            ary = parse("cxq[{:d}],q[{:d}];", instr)
            if ary is not None:
                cir.add_CNOT_gate(mapping[ary[0]], mapping[ary[1]])
        elif instr[0:2] == "cz":
            ary = parse("czq[{:d}],q[{:d}];", instr)
            cir.add_CZ_gate(mapping[ary[0]], mapping[ary[1]])
        elif instr[0:4] == "swap":
            ary = parse("swapq[{:d}],q[{:d}];", instr)
            cir.add_SWAP_gate(mapping[ary[0]], mapping[ary[1]])
        elif instr[0:2] == "id":
            ary = parse("idq[{:d}];", instr)
            cir.add_gate(Identity(mapping[ary[0]]))
        elif instr[0:2] == "xq":
            ary = parse("xq[{:d}];", instr)
            cir.add_X_gate(mapping[ary[0]])
        elif instr[0:2] == "yq":
            ary = parse("yq[{:d}];", instr)
            cir.add_Y_gate(mapping[ary[0]])
        elif instr[0:2] == "zq":
            ary = parse("zq[{:d}];", instr)
            cir.add_Z_gate(mapping[ary[0]])
        elif instr[0:2] == "hq":
            ary = parse("hq[{:d}];", instr)
            cir.add_H_gate(mapping[ary[0]])
        elif instr[0:2] == "sq":
            ary = parse("sq[{:d}];", instr)
            cir.add_S_gate(mapping[ary[0]])
        elif instr[0:4] == "sdgq":
            ary = parse("sdgq[{:d}];", instr)
            cir.add_Sdag_gate(mapping[ary[0]])
        elif instr[0:2] == "tq":
            ary = parse("tq[{:d}];", instr)
            cir.add_T_gate(mapping[ary[0]])
        elif instr[0:4] == "tdgq":
            ary = parse("tdgq[{:d}];", instr)
            cir.add_Tdag_gate(mapping[ary[0]])
        elif instr[0:2] == "rx":
            ary = parse("rx({:g})q[{:d}];", instr)
            cir.add_RX_gate(mapping[ary[1]], -ary[0])
        elif instr[0:2] == "ry":
            ary = parse("ry({:g})q[{:d}];", instr)
            cir.add_RY_gate(mapping[ary[1]], -ary[0])
        elif instr[0:2] == "rz":
            ary = parse("rz({:g})q[{:d}];", instr)
            cir.add_RZ_gate(mapping[ary[1]], -ary[0])
        elif instr[0:1] == "p":
            ary = parse("p({:g})q[{:d}];", instr)
            cir.add_U1_gate(mapping[ary[1]], ary[0])
        elif instr[0:2] == "u1":
            ary = parse("u1({:g})q[{:d}];", instr)
            cir.add_U1_gate(mapping[ary[1]], ary[0])
        elif instr[0:2] == "u2":
            ary = parse("u2({:g},{:g})q[{:d}];", instr)
            cir.add_U2_gate(mapping[ary[2]], ary[0], ary[1])
        elif instr[0:2] == "u3":
            ary = parse("u3({:g},{:g},{:g})q[{:d}];", instr)
            cir.add_U3_gate(mapping[ary[3]], ary[0], ary[1], ary[2])
        elif instr[0:1] == "u":
            ary = parse("u({:g},{:g},{:g})q[{:d}];", instr)
            if ary is not None:
                cir.add_U3_gate(mapping[ary[3]], ary[0], ary[1], ary[2])
        elif instr[0:3] == "sxq":
            ary = parse("sxq[{:d}];", instr)
            cir.add_sqrtX_gate(mapping[ary[0]])
        elif instr[0:5] == "sxdgq":
            ary = parse("sxdgq[{:d}];", instr)
            cir.add_sqrtXdag_gate(mapping[ary[0]])
        elif instr[0:5] == "cresq":
            ary = parse("CResq[{:d}],q[{:d}];", instr)
            cir.add_gate(CRes(mapping[ary[0]], mapping[ary[1]]))
        elif instr[0:8] == "cresdagq":
            ary = parse("CResdagq[{:d}],q[{:d}];", instr)
            cir.add_gate(CResdag(mapping[ary[0]], mapping[ary[1]]))
        elif instr[0:11] == "densematrix":
            ary = search("densematrix({:d},{:d}", instr)
            parsestr = "densematrix({:d},{:d}"
            for i in range(4 ** ary[0]):
                parsestr += ",{:g},{:g}"
            for i in range(ary[1]):
                parsestr += ",{:d}"
            parsestr += ")q[{:d}]"
            for i in range(ary[0] + ary[1] - 1):
                parsestr += ",q[{:d}]"
            parsestr += ";"
            deary = parse(parsestr, instr)
            gate_mat = np.zeros((2 ** ary[0], 2 ** ary[0]), dtype="complex")
            bas = 2
            for i in range(2 ** ary[0]):
                for j in range(2 ** ary[0]):
                    gate_mat[i][j] = deary[bas] + deary[bas + 1] * 1.0j
                    bas += 2
            control_values = []
            for i in range(ary[1]):
                control_values.append(mapping[deary[bas]])
                bas += 1
            terget_indexes = []
            for i in range(ary[0]):
                terget_indexes.append(mapping[deary[bas]])
                bas += 1

            dense_gate = DenseMatrix(terget_indexes, gate_mat)
            for i in range(ary[1]):
                control_index = deary[bas]
                bas += 1
                dense_gate.add_control_qubit(control_index, control_values[i])
            cir.add_gate(dense_gate)
        elif instr[0:7] == "measure":
            ary = parse("measure{q}->{c};", instr)
            if ary is not None:
                q = ary["q"].strip()
                c = ary["c"].strip()
                qary = parse("{s}[{:d}]", q)
                cary = parse("{s}[{:d}]", c)
                # print(f"qary: {qary}")
                # print(f"cary: {cary}")
                if qary is not None and cary is not None:
                    # print(f"qary[0]: {qary[0]}")
                    # print(f"cary[0]: {cary[0]}")
                    measurement_gate = Measurement(qary[0], cary[0])
                    # print(
                    #     f"measurement_gate.name: {measurement_gate.get_name()}")
                    # print(
                    #     f"measurement_gate get_target_index_list: {measurement_gate.get_target_index_list()}")
                    cir.add_gate(measurement_gate)
                else:
                    for i in range(n):
                        measurement_gate = Measurement(i, i)
                        cir.add_gate(measurement_gate)
        else:
            print("// unknown line:", instr)
    return cir


def state_to_strs(state: QuantumState) -> typing.List[str]:
    return state.to_string().split("\n")


def strs_to_state(strs: typing.List[str]) -> QuantumState:
    qubit = parse(" * Qubit Count : {:d}", strs[1])[0]
    Dim = 2**qubit
    state_vec: typing.List[complex] = [0.0 + 0.0j] * Dim
    for i in range(Dim):
        ary = search("({:g},{:g})", strs[4 + i])
        state_vec[i] = ary[0] + ary[1] * 1.0j
    state = QuantumState(qubit)
    state.load(state_vec)
    return state



from qulacs import NoiseSimulator, Observable, QuantumCircuit, QuantumState


def get_measurement(circuit: QuantumCircuit, obs: Observable, shots: int) -> float:
    """
    Args:
        circuit (qulacs.Quantumcircuit):
        obs (qulacs.Observable)
        shots (int):  number of samples for each observable
    Return:
        :float: sampled expectation value of the observable
    """
    n_term = obs.get_term_count()
    n_qubit = obs.get_qubit_count()

    exp = 0 + 0.0j

    for i in range(n_term):
        pauli_term = obs.get_term(i)
        coef = pauli_term.get_coef()
        pauli_id = pauli_term.get_pauli_id_list()
        pauli_index = pauli_term.get_index_list()
        if len(pauli_id) == 0:  # means identity
            exp += coef
            continue
        buf_state = QuantumState(n_qubit)
        measurement_circuit = circuit.copy()

        for single_pauli, index in zip(pauli_id, pauli_index):
            if single_pauli == 1:
                measurement_circuit.add_H_gate(index)
            elif single_pauli == 2:
                measurement_circuit.add_Sdag_gate(index)
                measurement_circuit.add_H_gate(index)
        measurement_circuit.update_quantum_state(buf_state)
        samples = buf_state.sampling(shots)
        mask = "".join(
            ["1" if n_qubit - 1 - k in pauli_index else "0" for k in range(n_qubit)]
        )
        masked = int(mask, 2)
        exp += (
            coef
            * sum(
                list(map(lambda x: int((-1) ** (bin(x & masked).count("1"))), samples))
            )
            / shots
        )

    return exp.real


def get_noise_meseurment(circuit: QuantumCircuit, obs: Observable, shots: int) -> float:
    """
    Args:
        circuit (qulacs.Quantumcircuit) (have noise)
        obs (qulacs.Observable)
        shots (int):  number of samples for each observable
    Return:
        :float: sampled expectation value of the observable

    this func use noisesimulator
    """
    n_term = obs.get_term_count()
    n_qubit = circuit.get_qubit_count()

    exp = 0 + 0.0j

    for i in range(n_term):
        pauli_term = obs.get_term(i)
        coef = pauli_term.get_coef()
        pauli_id = pauli_term.get_pauli_id_list()
        pauli_index = pauli_term.get_index_list()
        if len(pauli_id) == 0:  # means identity
            exp += coef
            continue
        buf_state = QuantumState(n_qubit)
        measurement_circuit = circuit.copy()

        for single_pauli, index in zip(pauli_id, pauli_index):
            if single_pauli == 1:
                measurement_circuit.add_H_gate(index)
            elif single_pauli == 2:
                measurement_circuit.add_Sdag_gate(index)
                measurement_circuit.add_H_gate(index)

        nsim = NoiseSimulator(measurement_circuit, buf_state)
        samples = nsim.execute(shots)
        mask = "".join(
            ["1" if n_qubit - 1 - k in pauli_index else "0" for k in range(n_qubit)]
        )
        masked = int(mask, 2)
        exp += (
            coef
            * sum(
                list(map(lambda x: int((-1) ** (bin(x & masked).count("1"))), samples))
            )
            / shots
        )

    return exp.real


def get_noiseevo_meseurment(
    circuit: QuantumCircuit, obs: Observable, shots: int
) -> float:
    """
    Args:
        circuit (qulacs.Quantumcircuit) (have noise)
        obs (qulacs.Observable)
        shots (int):  number of samples for each observable
    Return:
        :float: sampled expectation value of the observable

    this func NOT use noisesimulator
    """
    n_term = obs.get_term_count()
    n_qubit = circuit.get_qubit_count()

    exp = 0 + 0.0j

    for i in range(n_term):
        pauli_term = obs.get_term(i)
        coef = pauli_term.get_coef()
        pauli_id = pauli_term.get_pauli_id_list()
        pauli_index = pauli_term.get_index_list()
        if len(pauli_id) == 0:  # means identity
            exp += coef
            continue

        buf_state = QuantumState(n_qubit)
        measurement_circuit = circuit.copy()

        for single_pauli, index in zip(pauli_id, pauli_index):
            if single_pauli == 1:
                measurement_circuit.add_H_gate(index)
            elif single_pauli == 2:
                measurement_circuit.add_Sdag_gate(index)
                measurement_circuit.add_H_gate(index)
        mask = "".join(
            ["1" if n_qubit - 1 - k in pauli_index else "0" for k in range(n_qubit)]
        )
        masked = int(mask, 2)
        for _ in range(shots):
            buf_state.set_zero_state()
            measurement_circuit.update_quantum_state(buf_state)
            sample = buf_state.sampling(1)[0]
            exp += coef * int((-1) ** (bin(sample & masked).count("1"))) / shots

    return exp.real



import subprocess
from typing import List

import typer

from ouqu_tp.internal.getval import getval_noiseevo_do
from ouqu_tp.internal.sampleval import sampleval_noiseevo_do
from ouqu_tp.internal.simulate import simulate_noiseevo_do

app = typer.Typer()


@app.command("getval")
def getval_noisyevo_call(
    input_qasm_file: str = typer.Option(...),
    input_cnot_json_file: str = typer.Option(...),
    input_openfermion_file: str = typer.Option(...),
    shots: int = typer.Option(...),
    dt: float = 0.01,
    OZ: float = 1.0,
    OX: float = 1.0,
    ORes: float = 1.0,
    decay_rate_ph: float = 0.0,
    decay_rate_amp: float = 0.0,
    evodt: float = 0.5,
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(
                [
                    "staq",
                    "-S",
                    "-O2",
                    "-m",
                    "-d",
                    input_cnot_json_file,
                    "--evaluate-all",
                    input_qasm_file,
                ]
            )
            .decode()
            .splitlines()
        )
    print(
        getval_noiseevo_do(
            cpl_qasm,
            input_openfermion_file,
            shots,
            dt,
            OZ,
            OX,
            ORes,
            decay_rate_ph,
            decay_rate_amp,
            evodt,
        )
    )


@app.command("sampleval")
def sampleval_noisyevo_call(
    input_qasm_file: str = typer.Option(...),
    input_cnot_json_file: str = typer.Option(...),
    input_openfermion_file: str = typer.Option(...),
    shots: int = typer.Option(...),
    dt: float = 0.01,
    OZ: float = 1.0,
    OX: float = 1.0,
    ORes: float = 1.0,
    decay_rate_ph: float = 0.0,
    decay_rate_amp: float = 0.0,
    evodt: float = 0.5,
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(
                [
                    "staq",
                    "-S",
                    "-O2",
                    "-m",
                    "-d",
                    input_cnot_json_file,
                    "--evaluate-all",
                    input_qasm_file,
                ]
            )
            .decode()
            .splitlines()
        )
    print(
        sampleval_noiseevo_do(
            cpl_qasm,
            input_openfermion_file,
            shots,
            dt,
            OZ,
            OX,
            ORes,
            decay_rate_ph,
            decay_rate_amp,
            evodt,
        )
    )


@app.command("simulate")
def simulate_noisyevo_call(
    input_qasm_file: str = typer.Option(...),
    input_cnot_json_file: str = typer.Option(...),
    shots: int = typer.Option(...),
    dt: float = 0.01,
    OZ: float = 1.0,
    OX: float = 1.0,
    ORes: float = 1.0,
    decay_rate_ph: float = 0.0,
    decay_rate_amp: float = 0.0,
    evodt: float = 0.5,
    direct_qasm: bool = False,
) -> None:
    cpl_qasm: List[str]
    if direct_qasm:
        cpl_qasm = open(input_qasm_file, "r").readlines()
    else:
        cpl_qasm = (
            subprocess.check_output(
                [
                    "staq",
                    "-S",
                    "-O2",
                    "-m",
                    "-d",
                    input_cnot_json_file,
                    "--evaluate-all",
                    input_qasm_file,
                ]
            )
            .decode()
            .splitlines()
        )
    (kekka, n_qubit) = simulate_noiseevo_do(
        cpl_qasm,
        shots,
        dt,
        OZ,
        OX,
        ORes,
        decay_rate_ph,
        decay_rate_amp,
        evodt,
    )
    # input_listを直接ぶち込む
    for aaaa in kekka:
        moziretu = "{:b}".format(aaaa)
        while len(moziretu) < n_qubit:
            moziretu = "0" + moziretu
        print(moziretu)


if __name__ == "__main__":
    app()



import cmath
import typing
from cmath import atan, isclose, phase, pi, sqrt
from logging import NullHandler, getLogger
from typing import List, Tuple

import numpy as np
import numpy.typing as npt
import qulacs
from qulacs import QuantumCircuit
from qulacs.gate import RX, RZ, DenseMatrix, Identity, X, merge, sqrtX

logger = getLogger(__name__)
logger.addHandler(NullHandler())


def tran_ouqu_single(
    input_gate: qulacs.QuantumGateBase,
) -> typing.List[qulacs.QuantumGateBase]:
    # 1qubitのDenseMatrixゲートを入力し、 阪大のList[gate]の形に合わせます
    fugouZ = -1

    if len(input_gate.get_target_index_list()) != 1:
        logger.error("input gate is not single")
        raise RuntimeError("input gate is not single")
    if len(input_gate.get_control_index_list()) != 0:
        logger.error("input gate have control qubit")
        raise RuntimeError("input gate have control qubit")
    matrix = input_gate.get_matrix()
    qubit = input_gate.get_target_index_list()[0]

    out_gates: typing.List[qulacs.QuantumGateBase] = []
    # Rz単騎
    if cmath.isclose(abs(matrix[0][0]), 1, abs_tol=1e-5):
        degA = phase(matrix[1][1] / matrix[0][0]) * fugouZ
        if isclose(degA, 0):
            return out_gates
        out_gates.append(RZ(qubit, degA))
        return out_gates

    # Rz X
    if isclose(abs(matrix[0][0]), 0, abs_tol=1e-5):
        degA = phase(matrix[1][0] / matrix[0][1]) * fugouZ
        out_gates.append(RZ(qubit, degA))
        out_gates.append(X(qubit))
        return out_gates

    # Rz sqrtX Rz

    if isclose(abs(matrix[0][0]), cmath.sqrt(0.5), abs_tol=1e-5):
        degA = (phase(matrix[0][1] / matrix[0][0]) + pi / 2) * fugouZ
        degB = (phase(matrix[1][0] / matrix[0][0]) + pi / 2) * fugouZ
        out_gates.append(RZ(qubit, degA))
        out_gates.append(sqrtX(qubit))
        out_gates.append(RZ(qubit, degB))
        return out_gates

    # Rz sqrtX Rz sqrtX Rz
    adbc_mto = (matrix[0][0] * matrix[1][1]) / (matrix[0][1] * matrix[1][0])
    adbc = abs(adbc_mto)
    degB_com = -2 * atan(sqrt(adbc))  # 0～-π
    degB = degB_com.real * fugouZ
    degA = phase(-matrix[0][1] / matrix[0][0]) * fugouZ
    degC = phase(-matrix[1][0] / matrix[0][0]) * fugouZ

    out_gates.append(RZ(qubit, degA))
    out_gates.append(sqrtX(qubit))
    out_gates.append(RZ(qubit, degB))
    out_gates.append(sqrtX(qubit))
    out_gates.append(RZ(qubit, degC))
    return out_gates


def tran_ouqu_multi(inputcircuit: QuantumCircuit) -> QuantumCircuit:
    n_qubit = inputcircuit.get_qubit_count()
    anscircuit = QuantumCircuit(n_qubit)
    bitSingleGates = []

    for i in range(n_qubit):
        bitSingleGates.append(Identity(i))
    gate_num = inputcircuit.get_gate_count()
    for i in range(gate_num):
        ingate = inputcircuit.get_gate(i)

        # 測定はスキップ
        if ingate.get_name() == "CPTP" or ingate.get_name() == "Instrument":
            # TODO: 測定の追加位置（bitSingleGatesはまとめているためここで測定を追加できない）
            # anscircuit.add_gate(ingate)
            continue

        if (
            len(ingate.get_control_index_list()) + len(ingate.get_target_index_list())
            <= 1
        ):
            target = ingate.get_target_index_list()[0]
            newgate = merge(bitSingleGates[target], ingate)
            bitSingleGates[target] = newgate
        else:
            bits = ingate.get_control_index_list() + ingate.get_target_index_list()
            for i in bits:
                for tuigate in tran_ouqu_single(bitSingleGates[i]):
                    anscircuit.add_gate(tuigate)
                bitSingleGates[i] = Identity(i)
            anscircuit.add_gate(ingate)

    for i in range(n_qubit):
        for tuigate in tran_ouqu_single(bitSingleGates[i]):
            anscircuit.add_gate(tuigate)

    # 測定の追加
    for i in range(gate_num):
        ingate = inputcircuit.get_gate(i)
        if ingate.get_name() == "CPTP" or ingate.get_name() == "Instrument":
            # print(f"tran_ouqu_multi ingate: {ingate}")
            # print(
            #     f"tran_ouqu_multi get_target_index_list: {ingate.get_target_index_list()}")
            anscircuit.add_gate(ingate)

    return anscircuit


def CRes(targetA: int, targetB: int) -> qulacs.QuantumGateBase:
    gate_mat = np.array(
        [[1, 0, -1.0j, 0], [0, 1, 0, 1.0j], [-1.0j, 0, 1, 0], [0, 1.0j, 0, 1]]
    )
    return DenseMatrix([targetA, targetB], gate_mat / sqrt(2))  # type:ignore


def CResdag(targetA: int, targetB: int) -> qulacs.QuantumGateBase:
    gate_mat = np.array(
        [[1, 0, 1.0j, 0], [0, 1, 0, -1.0j], [1.0j, 0, 1, 0], [0, -1.0j, 0, 1]]
    )
    return DenseMatrix([targetA, targetB], gate_mat / sqrt(2))  # type:ignore


def CNOT_to_CRes(inputcircuit: QuantumCircuit) -> QuantumCircuit:
    n_qubit = inputcircuit.get_qubit_count()
    # 元のゲートにCNOTゲートが入っていたら、CResゲートに変換する
    anscircuit = QuantumCircuit(n_qubit)
    gate_num = inputcircuit.get_gate_count()
    for i in range(gate_num):
        ingate = inputcircuit.get_gate(i)
        if ingate.get_name() == "CNOT":
            target = ingate.get_target_index_list()[0]
            control = ingate.get_control_index_list()[0]
            anscircuit.add_gate(RX(target, pi / 2))
            anscircuit.add_gate(CRes(control, target))
            anscircuit.add_gate(RZ(control, pi / 2))
        else:
            anscircuit.add_gate(ingate)
    return anscircuit


def check_is_CRes(ingate: qulacs.QuantumGateBase) -> bool:
    if not ingate.get_name() == "DenseMatrix":
        return False
    if len(ingate.get_target_index_list()) != 2:
        return False
    if len(ingate.get_control_index_list()) != 0:
        return False

    true_mat = np.array(
        [[1, 0, -1.0j, 0], [0, 1, 0, 1.0j], [-1.0j, 0, 1, 0], [0, 1.0j, 0, 1]]
    ) / sqrt(2)
    return np.allclose(true_mat, ingate.get_matrix())


def check_is_CResdag(ingate: qulacs.QuantumGateBase) -> bool:
    if not ingate.get_name() == "DenseMatrix":
        return False
    if len(ingate.get_target_index_list()) != 2:
        return False
    if len(ingate.get_control_index_list()) != 2:
        return False

    true_mat = np.array(
        [[1, 0, 1.0j, 0], [0, 1, 0, -1.0j], [1.0j, 0, 1, 0], [0, -1.0j, 0, 1]]
    ) / sqrt(2)
    return np.allclose(true_mat, ingate.get_matrix())


def tran_to_pulse_tyukan(
    inputcircuit_in: QuantumCircuit,
    Res_list: List[Tuple[int, int]],
    dt: float,
    OZ: float,
    OX: float,
    ORes: float,
    mergin: int = 0,
) -> Tuple[List[Tuple[int, int, int]], int]:
    RZome = dt * OZ
    RXome = dt * OX
    CResome = dt * ORes
    inputcircuit = inputcircuit_in.copy()
    n_qubit = inputcircuit.get_qubit_count()

    inputcircuit = CNOT_to_CRes(inputcircuit)
    inputcircuit = tran_ouqu_multi(inputcircuit)
    """
    「任意の1qubit回転 + CNOT,CRes 」でできたcircuitを、中間表現に直します。
    各パルスは、(ゲート番号、スタート時刻、ゲート長さ) を持ちます。
    ゲート長さは、回転角/omeです。

    「空白」を表すゲートも入れる、　(何もしない間にもデコヒーレンスは発生するので)
    ゲート番号は、ZZZZZXXXXXRRRRR...RRSSSSS のような定義をされる
    (Sは空白ゲートです)

    この段階で,ゲートの向きをqulacsではなく、標準基準に戻します
    """
    logger.debug(n_qubit)
    logger.debug(inputcircuit)
    bangou = np.zeros((n_qubit, n_qubit), int)
    for i in range(n_qubit):
        for j in range(n_qubit):
            bangou[i][j] = -1
    for i in range(len(Res_list)):
        (ppp, qqq) = Res_list[i]
        bangou[ppp][qqq] = i

    saigo_zikan = np.zeros(n_qubit, int)
    pulse_comp: List[Tuple[int, int, int]] = []
    gate_num = inputcircuit.get_gate_count()

    space_ban_start = n_qubit * 2 + len(Res_list)

    for i in range(gate_num):
        ingate = inputcircuit.get_gate(i)

        target = ingate.get_target_index_list()[0]
        if ingate.get_name() == "Z-rotation":
            matrix = ingate.get_matrix()
            angle = phase(matrix[1][1] / matrix[0][0])
            if angle < -1e-6:
                angle += pi * 2
            pulse_kaz = int(angle / (RZome * 2) + 0.5)
            if pulse_kaz > 0:
                pulse_comp.append((target, saigo_zikan[target], pulse_kaz))
                saigo_zikan[target] += pulse_kaz + mergin

        elif ingate.get_name() == "sqrtX":
            pulse_kaz = int(pi / 2 / (RXome * 2) + 0.5)
            pulse_comp.append((target + n_qubit, saigo_zikan[target], pulse_kaz))
            saigo_zikan[target] += pulse_kaz + mergin
        elif ingate.get_name() == "X":
            pulse_kaz = int(pi / (RXome * 2) + 0.5)
            pulse_comp.append((target + n_qubit, saigo_zikan[target], pulse_kaz))
            saigo_zikan[target] += pulse_kaz + mergin
        elif check_is_CRes(ingate):
            control = ingate.get_target_index_list()[0]
            target = ingate.get_target_index_list()[1]
            ban = bangou[control][target]
            if ban == -1:
                logger.error(f"({control},{target}) gate is not in Res_list")
            start = max(saigo_zikan[target], saigo_zikan[control])
            if saigo_zikan[target] < start:
                pulse_comp.append(
                    (
                        space_ban_start + target,
                        saigo_zikan[target],
                        start - saigo_zikan[target],
                    )
                )
            if saigo_zikan[control] < start:
                pulse_comp.append(
                    (
                        space_ban_start + control,
                        saigo_zikan[control],
                        start - saigo_zikan[control],
                    )
                )
            pulse_kaz = int(pi / (CResome * 4) + 0.5)
            pulse_comp.append((ban + n_qubit * 2, start, pulse_kaz))
            saigo_zikan[target] = start + pulse_kaz + mergin
            saigo_zikan[control] = start + pulse_kaz + mergin
        else:
            logger.error("this gate is not (RZ,sx,x,CRes)")
            logger.error(ingate)
            raise RuntimeError("this gate is not (RZ,sx,x,CRes)")
    for aaa in pulse_comp:
        logger.debug(aaa)

    T = int(np.amax(saigo_zikan))

    return (pulse_comp, T)


def tran_to_pulse(
    inputcircuit: QuantumCircuit,
    Res_list: List[Tuple[int, int]],
    dt: float,
    OZ: float,
    OX: float,
    ORes: float,
    mergin: int = 0,
) -> npt.NDArray[np.float64]:
    RZome = dt * OZ
    RXome = dt * OX
    CResome = dt * ORes

    (pulse_comp, T) = tran_to_pulse_tyukan(
        inputcircuit, Res_list, dt, OZ, OX, ORes, mergin
    )
    """
    ゲートパルス中間表現を関数内で取得し、それをnumpy array に直します。
    arrayは、[ゲート名+1][時間(1パルス単位)] の配列です。

    array[0][時間(1パルス単位)]は、 実時間の配列です。{0,dt,dt*2,dt*3...}
    """
    n_qubit = inputcircuit.get_qubit_count()

    result_pulse = np.zeros((n_qubit * 2 + len(Res_list) + 1, int(T)))

    for ple in pulse_comp:
        (gate_ban, start, time) = ple
        if gate_ban >= n_qubit * 2 + len(Res_list):
            continue  # 空白に対応するので
        omega = RZome
        if gate_ban >= n_qubit:
            omega = RXome
        if gate_ban >= n_qubit * 2:
            omega = CResome
        for j in range(start, time + start):
            result_pulse[gate_ban + 1][j] = omega
    for j in range(T):
        result_pulse[0][j] = dt * j
    return result_pulse


def pulse_to_circuit(
    n_qubit: int,
    pulse_array: npt.NDArray[np.float64],
    Res_list: List[Tuple[int, int]],
) -> QuantumCircuit:
    # パルス情報が与えられたとき、量子回路を実行する関数です
    anscircuit = QuantumCircuit(n_qubit)
    m_kaz = n_qubit * 2 + len(Res_list)
    renzoku = np.zeros(m_kaz)
    T = len(pulse_array[0])
    for i in range(T + 1):
        for j in range(m_kaz):
            if i < T and pulse_array[j + 1][i] > 1e-8:
                renzoku[j] += pulse_array[j + 1][i]
            elif renzoku[j] > 1e-8:
                if j < n_qubit:
                    # RZ gate
                    anscircuit.add_gate(RZ(j, -renzoku[j] * 2))
                elif j < n_qubit * 2:
                    anscircuit.add_gate(RX(j - n_qubit, -renzoku[j] * 2))
                else:
                    (control, target) = Res_list[j - n_qubit * 2]
                    anscircuit.add_gate(CRes(control, target))
                renzoku[j] = 0
    return anscircuit



from typing import List, Tuple

from qulacs import GeneralQuantumOperator, Observable, QuantumCircuit
from qulacs.gate import (
    DepolarizingNoise,
    NoisyEvolution_fast,
    TwoQubitDepolarizingNoise,
)

from ouqu_tp.internal.tran import tran_to_pulse_tyukan


def auto_noise(
    inputcircuit_in: QuantumCircuit,
    p1: float,
    p2: float,
    pm: float,
    pp: float,
) -> QuantumCircuit:
    inputcircuit = inputcircuit_in.copy()
    n_qubit = inputcircuit.get_qubit_count()
    testcircuit = QuantumCircuit(n_qubit)
    for i in range(n_qubit):
        testcircuit.add_gate(DepolarizingNoise(i, pp))
    gate_num = inputcircuit.get_gate_count()
    for i in range(gate_num):
        ingate = inputcircuit.get_gate(i)
        testcircuit.add_gate(ingate)

        gate_index_list = (
            ingate.get_control_index_list() + ingate.get_target_index_list()
        )

        if len(gate_index_list) == 1:
            testcircuit.add_gate(DepolarizingNoise(gate_index_list[0], p1))
        if len(gate_index_list) == 2:
            testcircuit.add_gate(
                TwoQubitDepolarizingNoise(gate_index_list[0], gate_index_list[1], p2)
            )

        if len(gate_index_list) > 2:
            # 本来ありえません　3ビット以上のやつは
            raise RuntimeError("3ビット以上のやつを与えないでください もしくは何かバグがあるので、連絡して下さい")

    for i in range(n_qubit):
        testcircuit.add_gate(DepolarizingNoise(i, pm))
    return testcircuit


def auto_Res_list(inputcircuit: QuantumCircuit) -> List[Tuple[int, int]]:
    Res_list: List[Tuple[int, int]] = []

    gate_num = inputcircuit.get_gate_count()
    Res_set = set(Res_list)
    for i in range(gate_num):
        ingate = inputcircuit.get_gate(i)

        gate_index_list = (
            ingate.get_control_index_list() + ingate.get_target_index_list()
        )
        if len(gate_index_list) == 2:
            gen_Res_pair = (gate_index_list[0], gate_index_list[1])
            if gen_Res_pair not in Res_set:
                Res_set.add(gen_Res_pair)
                Res_list.append(gen_Res_pair)

    return Res_list


def auto_evo_noise(
    inputcircuit_in: QuantumCircuit,
    dt: float,
    OZ: float,
    OX: float,
    ORes: float,
    decay_rate_ph: float,
    decay_rate_amp: float,
    evodt: float,
) -> QuantumCircuit:
    """
    :param pulce_comp:  各パルスは、(ゲート番号、スタート時刻、ゲート長さ)で、　tran_to_pulse_tyukanの帰り値を渡すことが多い
    :param dt: 1パルスの時間
    :param OZ: 1時間あたりの回転量(not 1パルス)
    :param OX: 1時間あたりの回転量
    :param ORes: 1時間あたりの回転量
    :param decay_rate_ph:1時間あたりの位相のデコヒーレンス量?
    :param decay_rate_amp:1時間あたりのampのデコヒーレンス量?
    :param n_qubit: キュービット数
    :param evodt: NoisyEvolutionでの1ステップの時間 これが小さいほど正確だけど計算負荷かかる
    """
    inputcircuit = inputcircuit_in.copy()
    Res_list = auto_Res_list(inputcircuit)
    (pulse_comp, unuse) = tran_to_pulse_tyukan(inputcircuit, Res_list, dt, OZ, OX, ORes)
    n_qubit = inputcircuit.get_qubit_count()
    anscircuit = QuantumCircuit(n_qubit)
    for ple in pulse_comp:
        (gate_ban, start, time) = ple
        # print(gate_ban, start, time)
        if gate_ban < n_qubit * 2 or n_qubit * 2 + len(Res_list) <= gate_ban:
            if gate_ban < n_qubit:  # Z gate
                target = gate_ban
                hamiltonian = Observable(n_qubit)
                hamiltonian.add_operator(OZ, "Z {0}".format(target))
            elif gate_ban < n_qubit * 2:  # X gate
                target = gate_ban - n_qubit
                hamiltonian = Observable(n_qubit)
                hamiltonian.add_operator(OX, "X {0}".format(target))
            elif gate_ban >= n_qubit * 2 + len(Res_list):  # I gate
                target = gate_ban - (n_qubit * 2 + len(Res_list))
                hamiltonian = Observable(n_qubit)
            jump_op_list = [GeneralQuantumOperator(n_qubit) for i in range(2)]
            jump_op_list[0].add_operator(decay_rate_ph, "Z {0}".format(target))
            jump_op_list[1].add_operator(
                decay_rate_amp / 2 * 1j, "Y {0}".format(target)
            )
            jump_op_list[1].add_operator(decay_rate_amp / 2, "X {0}".format(target))

        else:
            (control, target) = Res_list[gate_ban - n_qubit * 2]
            hamiltonian = Observable(n_qubit)
            hamiltonian.add_operator(ORes, "X {0} Z {1}".format(target, control))
            jump_op_list = [GeneralQuantumOperator(n_qubit) for i in range(4)]
            jump_op_list[0].add_operator(decay_rate_ph, "Z {0}".format(target))
            jump_op_list[1].add_operator(
                decay_rate_amp / 2 * 1j, "Y {0}".format(target)
            )
            jump_op_list[1].add_operator(decay_rate_amp / 2, "X {0}".format(target))
            jump_op_list[2].add_operator(decay_rate_ph, "Z {0}".format(control))
            jump_op_list[3].add_operator(
                decay_rate_amp / 2 * 1j, "Y {0}".format(control)
            )
            jump_op_list[3].add_operator(decay_rate_amp / 2, "X {0}".format(control))

        # print(gate_ban,n_qubit,len(Res_list))
        # print(hamiltonian, start, time, evodt)
        opr = NoisyEvolution_fast(hamiltonian, jump_op_list, time * dt)
        anscircuit.add_gate(opr)

    return anscircuit



import typing


def input_strings() -> typing.List[str]:
    input_strs = []
    while True:
        try:
            line = input()
            if line == "END":  # END という3文字を受け取る
                break
            input_strs.append(line)
        except EOFError:
            break
    return input_strs



from typing import List, Tuple


def make_Cnet_put(input_strs: List[str]) -> None:
    qubit_num = int(input_strs[1])
    # connect_num = int(input_strs[2]) (not used)
    con = []

    print("{")
    print('  "couplings": [')
    for i in range(3, len(input_strs)):
        kazstr = input_strs[i].rstrip("\n").split(",")
        con.append([int(kazstr[0]), int(kazstr[1])])
        print("    {")
        print('      "control": ' + kazstr[0] + ",")
        print('      "target": ' + kazstr[1])
        if i + 1 < len(input_strs):
            print("    },")
        else:
            print("    }")
    print("  ],")
    print('  "name": "' + input_strs[0].rstrip("\n") + '",')
    print('  "qubits": [')
    for i in range(qubit_num):
        print("    {")
        print('      "id": ' + str(i))
        if i + 1 < qubit_num:
            print("    },")
        else:
            print("    }")
    print("  ]")
    print("}")


def get_connect(input_strs: List[str]) -> List[Tuple[int, int]]:
    con = []
    for i in range(3, len(input_strs)):
        kazstr = input_strs[i].rstrip("\n").split(",")
        con.append((int(kazstr[0]), int(kazstr[1])))
    return con



from qulacs import QuantumCircuit, QuantumState
from qulacs.gate import H
from qulacs.state import inner_product


def check_circuit(
    cirA: QuantumCircuit, cirB: QuantumCircuit, ok_rate: float = 0.999
) -> None:
    # ランダムなstateで6回試して、二つのcircuitが同じものかどうか確かめます。
    for i in range(6):
        stateA = QuantumState(cirA.get_qubit_count())
        if i > 0:
            stateA.set_Haar_random_state(i)
        else:
            H(0).update_quantum_state(stateA)
            H(1).update_quantum_state(stateA)

        stateB = stateA.copy()
        print(stateA)
        print(stateB)

        cirA.update_quantum_state(stateA)
        cirB.update_quantum_state(stateB)

        print(stateA)
        print(stateB)
        print("end")
        assert abs(inner_product(stateA, stateB)) > ok_rate
    return



from typing import List

import numpy as np
import numpy.typing as npt

from ouqu_tp.internal.make_Cnet import get_connect
from ouqu_tp.internal.QASMtoqulacs import QASM_to_qulacs, qulacs_to_QASM
from ouqu_tp.internal.tran import CNOT_to_CRes, tran_ouqu_multi, tran_to_pulse


def trance_do(input_strs: List[str]) -> List[str]:
    mtocircuit = QASM_to_qulacs(input_strs, remap_remove=False)
    trancircuit = tran_ouqu_multi(mtocircuit)
    return qulacs_to_QASM(trancircuit)


def trance_res_do(input_strs: List[str]) -> List[str]:
    mtocircuit = QASM_to_qulacs(input_strs, remap_remove=False)
    mtocircuit = CNOT_to_CRes(mtocircuit)
    trancircuit = tran_ouqu_multi(mtocircuit)
    return qulacs_to_QASM(trancircuit)


def trance_pulse_do(
    input_strs: List[str],
    Cnet_list: List[str],
    dt: float,
    OZ: float,
    OX: float,
    ORes: float,
    mergen: int = 0,
) -> npt.NDArray[np.float64]:
    mtocircuit = QASM_to_qulacs(input_strs, remap_remove=False)
    Res_list = get_connect(Cnet_list)
    result_array = tran_to_pulse(mtocircuit, Res_list, dt, OZ, OX, ORes, mergen)
    return result_array
